/**
 * The Splitter EIP pattern
 */

package org.ow2.petals.se.eip.patterns;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.net.URI;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import javax.activation.DataHandler;
import javax.activation.DataSource;
import javax.jbi.messaging.Fault;
import javax.jbi.messaging.MessagingException;
import javax.jbi.messaging.NormalizedMessage;
import javax.xml.transform.Source;
import javax.xml.transform.dom.DOMSource;
import javax.xml.xpath.XPathExpressionException;

import org.ow2.petals.component.framework.Constants.MEPConstants;
import org.ow2.petals.component.framework.api.exception.PEtALSCDKException;
import org.ow2.petals.component.framework.api.message.Exchange;
import org.ow2.petals.component.framework.jbidescriptor.generated.Consumes;
import org.ow2.petals.component.framework.util.SourceHelper;
import org.ow2.petals.component.framework.util.XMLUtil;
import org.ow2.petals.se.eip.EIPConstants;
import org.ow2.petals.se.eip.EIPException;
import org.ow2.petals.se.eip.ExchangeContext;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

/**
 * @author Adrien LOUIS - EBMWebSourcing
 * @author Frederic GARDES
 * 
 */
/**
 * @author fgardes
 * 
 */
public class Splitter extends AbstractPattern {

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.se.eip.patterns.AbstractPattern#getPatternName()
     */
    @Override
    protected String getPatternName() {
        return "Splitter";
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.se.eip.patterns.AbstractPattern#process(org.ow2.petals
     * .component.framework.api.message.Exchange,
     * org.ow2.petals.se.eip.ExchangeContext)
     */
    @Override
    public void process(Exchange originalExchange, ExchangeContext context)
            throws MessagingException, PEtALSCDKException {
        List<Consumes> consumesList = null;
        Consumes consumes = null;
        Iterator<DataHandler> iterator = null;
        Document inMessageContent = null;

        if (!originalExchange.isDoneStatus()) {
            consumesList = context.getSUConsumes(originalExchange.getEndpoint());
            if (consumesList == null || consumesList.size() != 1) {
                throw new PEtALSCDKException(
                        "The Service Unit must define only one consumes section");
            }
            consumes = consumesList.get(0);
            if (consumes == null) {
                throw new PEtALSCDKException(
                        "The Service Unit must define a valid consumes section");
            }

            // retrieve the document only once to avoid the Streaming issue
            try {
                List<Source> list = null;
                inMessageContent = originalExchange.getInMessageContentAsDocument();
                if (!isAttachmentMode(context.getExtensions())) {
                    String splitterPath = null;
                    splitterPath = getPath(context.getExtensions());
                    if (splitterPath == null || splitterPath.length() == 0) {
                        throw new PEtALSCDKException("The Service Unit must define a path section");
                    }
                    logger.finest("path:" + splitterPath);
                    NodeList nodeList = PatternHelper.evaluateXPathExpression(splitterPath,
                            inMessageContent);
                    if (nodeList == null) {
                        throw new PEtALSCDKException(
                                "The Service Unit cannot define the list of nodes to send");
                    }
                    logger.finest("node list generated with " + nodeList.getLength() + " elements");
                    list = new ArrayList<Source>(nodeList.getLength());
                    for (int i = 0; i < nodeList.getLength(); i++) {
                        list.add(new DOMSource(nodeList.item(i)));
                    }
                    if (list == null) {
                        throw new PEtALSCDKException(
                                "The Service Unit cannot define the list of sources to send");
                    }
                    logger.finest("source list generated with " + list.size() + " elements");
                    sendNodeList(inMessageContent, originalExchange, context, consumes, list, true,
                            true);
                } else if (!originalExchange.getInMessageAttachments().isEmpty()) {
                    list = new ArrayList<Source>(originalExchange.getInMessageAttachments().size());
                    iterator = originalExchange.getInMessageAttachments().iterator();
                    while (iterator.hasNext()) {
                        DataHandler dataHandler = iterator.next();
                        String name = dataHandler.getName();
                        String type = dataHandler.getContentType();
                        logger.finest("type of the current attachment:" + type);
                        System.out.print("Attachment " + name + " has content type " + type);
                        if (type.equals("application/xml")) {
                            DataSource content = (DataSource) dataHandler.getContent();
                            list.add(SourceHelper
                                    .createSourceFromString(convertStreamToString(content
                                            .getInputStream())));
                        }
                    }
                    if (list == null) {
                        throw new PEtALSCDKException(
                                "The Service Unit cannot define the list of sources to send");
                    }
                    logger.finest("source list generated with " + list.size() + " elements");
                    sendNodeList(inMessageContent, originalExchange, context, consumes, list,
                            false, true);
                }

            } catch (XPathExpressionException e) {
                throw new PEtALSCDKException(e);
            } catch (InterruptedException e) {
                originalExchange.setError(e);
                originalExchange.setErrorStatus();
            } catch (IOException e) {
                throw new PEtALSCDKException(e);
            }

        }
    }

    private String convertStreamToString(InputStream is) {
        /*
         * To convert the InputStream to String we use the
         * BufferedReader.readLine() method. We iterate until the BufferedReader
         * return null which means there's no more data to read. Each line will
         * appended to a StringBuilder and returned as String.
         */
        BufferedReader reader = new BufferedReader(new InputStreamReader(is));
        StringBuilder sb = new StringBuilder();

        String line = null;
        try {
            while ((line = reader.readLine()) != null) {
                sb.append(line);
            }
        } catch (IOException e) {
            logger.fine(e.getMessage());
        } finally {
            try {
                is.close();
            } catch (IOException e) {
                logger.fine(e.getMessage());
            }
        }

        return sb.toString();
    }

    private void sendNodeList(Document inMessageContent, Exchange originalExchange,
            ExchangeContext context, Consumes consumes, List<Source> list, boolean copyAttachments,
            boolean copyProperties) throws MessagingException, PEtALSCDKException,
            InterruptedException {
        List<Exchange> splitExchangeList;
        NormalizedMessage receivedMessage = null;
        boolean fault = false;
        List<Source> receivedSourceList = null;
        boolean exception = false;

        splitExchangeList = new ArrayList<Exchange>(list.size());
        logger.finest("split exchange list generated");
        int nbTransaction = Math.abs((list.size() / EIPConstants.TRANSACTION_SIZE) + 1);

        for (int i = 0; i < nbTransaction; i++) {
            int begin = i * EIPConstants.TRANSACTION_SIZE;
            int end = Math.min(begin + EIPConstants.TRANSACTION_SIZE - 1, list.size() - 1);
            logger.finest("treatment of the transaction " + i + ": from " + begin + " to " + end);
            for (int j = begin; j <= end; j++) {
                Exchange splitExchange = PatternHelper.createConsumeExchange(consumes, context);
                if (splitExchange.getOperation() == null) {
                    splitExchange.setOperation(originalExchange.getOperation());
                }
                PatternHelper.copy(list.get(j), originalExchange.getInMessage(), splitExchange
                        .getInMessage(), copyAttachments, copyProperties);
                logSend(consumes, splitExchange);
                context.sendAsync(splitExchange);
                splitExchangeList.add(splitExchange);
                logger.finest("split exchange " + splitExchange.getExchangeId() + " sent");
            }
            receivedSourceList = new ArrayList<Source>(list.size());
            logger.finest("out message list generated");
            for (int j = begin; j <= end && !fault && !exception; j++) {
                // process the response
                try {
                    receivedMessage = PatternHelper.processSentExchange(context
                            .accept(splitExchangeList.get(j)), context);
                    logger.finest("split exchange analyzed");
                } catch (EIPException ex) {
                    if (!isExceptionRobust(context.getExtensions())) {
                        originalExchange.setError((Exception) ex.getCause());
                        originalExchange.setErrorStatus();
                        exception = true;
                        logger.finest("an exception is thrown");
                    } else {
                        Document document = createException();
                        document.setNodeValue(ex.getCause().getMessage());
                        receivedSourceList.add(SourceHelper.createSource(document));
                        logger.finest("an exception is saved");
                    }
                }
                if (receivedMessage != null) {
                    if (receivedMessage instanceof Fault && !isFaultRobust(context.getExtensions())) {
                        originalExchange.setFault((Fault) receivedMessage);
                        fault = true;
                        logger.finest("a fault is thrown");
                    } else {
                        receivedSourceList.add(receivedMessage.getContent());
                        logger.finest("a message is saved");
                    }
                    // put each properties
                    for (Object name : receivedMessage.getPropertyNames()) {
                        originalExchange.getOutMessage().setProperty((String) name,
                                receivedMessage.getProperty((String) name));
                    }
                    // put each attachments
                    for (Object name : receivedMessage.getAttachmentNames()) {
                        originalExchange.getOutMessage().addAttachment((String) name,
                                receivedMessage.getAttachment((String) name));
                    }
                }
            }
        }
        // reset the content of the in Message to avoid Streaming issue
        // TODO delete this when the Transporter will be corrected : the
        // JBI specification allow to have an in message null
        originalExchange.getInMessage().setContent(new DOMSource(inMessageContent));
        // manage out , fault and done differently for each mep
        if (!fault && !exception) {
            if (isWithOutPattern(originalExchange.getExchangePattern())) {
                // put out message to content
                originalExchange.getOutMessage().setContent(
                        concatNormalizedMesssageContent(receivedSourceList));
            } else {
                originalExchange.setDoneStatus();
            }
        }
    }

    /**
     * get the return capability from the MEP
     * 
     * @return true if the MEP is InOut or InOptionalOut, false else
     */
    private boolean isWithOutPattern(URI mep) {
        return MEPConstants.IN_OUT_PATTERN.equals(mep)
                || MEPConstants.IN_OPTIONAL_OUT_PATTERN.equals(mep);
    }

    /**
     * Create a document with the result node provided by the context
     * 
     * @return a Document with the result node provided by the context
     * @throws PEtALSCDKException
     */
    private Document createResult() throws PEtALSCDKException {

        Document document = XMLUtil.createDocument();
        Element rootElement = document.createElementNS(EIPConstants.NAMESPACE,
                EIPConstants.RESULT_ELEMENT);
        document.appendChild(rootElement);
        return document;
    }

    /**
     * Create a document with the exception node provided by the context
     * 
     * @return a Document with the exception node provided by the context
     * @throws PEtALSCDKException
     */
    private Document createException() throws PEtALSCDKException {

        Document document = XMLUtil.createDocument();
        Element rootElement = document.createElementNS(EIPConstants.NAMESPACE,
                EIPConstants.EXCEPTION_ELEMENT);
        document.appendChild(rootElement);
        return document;
    }

    /**
     * Build the aggregate source from a list of normalized messages with the
     * root node provided by the context.
     * 
     * @param normalizedMessageList
     *            The List of NormalizedMessage used
     * @return a Source with the content of each valid message content
     * @throws PEtALSCDKException
     */
    private Source concatNormalizedMesssageContent(List<Source> sourceList)
            throws PEtALSCDKException {
        Source source = null;
        Document aggregateDocument = createResult();
        if (sourceList != null) {
            for (int i = 0; i < sourceList.size(); i++) {
                source = sourceList.get(i);
                Document outDocument = SourceHelper.createDocument(source);
                Node node = aggregateDocument.adoptNode(outDocument.getDocumentElement());
                aggregateDocument.getDocumentElement().appendChild(node);
            }
        }
        return SourceHelper.createSource(aggregateDocument);
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.se.eip.patterns.AbstractPattern#validateMEP(java.net.URI)
     */
    @Override
    protected boolean validateMEP(URI mep) {
        return true;
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.se.eip.patterns.Pattern#init()
     */
    public void init() {
    }
}
